import tempfile
from pathlib import Path

import cbgen


class BGENSuite:
    timeout = 10 * 60.0

    def __init__(self):
        self._filepath = cbgen.example.get("merged_487400x220000.bgen")
        self._mfilepath = Path("metafile")
        with cbgen.bgen_file(self._filepath) as bgen:
            bgen.create_metafile(self._mfilepath, verbose=False)

    def setup(self):
        pass

    def time_bgen_file(self):
        with cbgen.bgen_file(self._filepath):
            pass

    def time_create_metafile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with cbgen.bgen_file(self._filepath) as bgen:
                bgen.create_metafile(Path(tmpdir) / "metafile", verbose=False)

    def time_bgen_metafile(self):
        with cbgen.bgen_metafile(self._mfilepath):
            pass

    def time_read_partitions(self):
        with cbgen.bgen_metafile(self._mfilepath) as mf:
            for i in range(mf.npartitions):
                mf.read_partition(i)



from dataclasses import dataclass
from typing import Any

__all__ = ["CData", "DtypeLike", "Variants", "Genotype", "Partition"]

# Waiting for official type hint: https://foss.heptapod.net/pypy/cffi/issues/456
CData = Any

# Waiting for numpy release: https://github.com/numpy/numpy/labels/static%20typing
DtypeLike = Any


@dataclass
class Genotype:
    """
    Genotype.

    >>> import cbgen
    >>>
    >>> bgen = cbgen.bgen_file(cbgen.example.get("haplotypes.bgen"))
    >>> mf = cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile"))
    >>> part = mf.read_partition(0)
    >>> gt = bgen.read_genotype(part.variants.offset[0])
    >>> print(type(gt))
    <class 'cbgen.typing.Genotype'>
    >>> print(gt.probability)
    [[1. 0. 1. 0.]
     [0. 1. 1. 0.]
     [1. 0. 0. 1.]
     [0. 1. 0. 1.]]
    >>> print(gt.phased)
    True
    >>> print(gt.ploidy)
    [2 2 2 2]
    >>> print(gt.missing)
    [False False False False]
    >>> mf.close()
    >>> bgen.close()

    Attributes
    ----------
    Probability
        Probability.
    phased
         Phasedness.
    ploidy
        Ploidy.
    missing
        Missingness.
    """

    probability: DtypeLike
    phased: DtypeLike
    ploidy: DtypeLike
    missing: DtypeLike


@dataclass
class Variants:
    """
    Variants.

    >>> import cbgen
    >>>
    >>> bgen = cbgen.bgen_file(cbgen.example.get("haplotypes.bgen"))
    >>> mf = cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile"))
    >>> part = mf.read_partition(0)
    >>> variants = part.variants
    >>> print(type(variants))
    <class 'cbgen.typing.Variants'>
    >>> print(variants.size)
    4
    >>> print(variants.id[3])
    b'SNP4'
    >>> print(variants.rsid[3])
    b'RS4'
    >>> print(variants.chromosome[3])
    b'1'
    >>> print(variants.position[3])
    4
    >>> print(variants.nalleles[3])
    2
    >>> print(variants.allele_ids[3])
    b'A,G'
    >>> print(variants.offset[3])
    273
    >>> mf.close()
    >>> bgen.close()

    Attributes
    ----------
    id
        Identification.
    rsid
         Reference SNP cluster ID.
    chromosome
        Chromosome.
    position
        Position.
    nalleles
        Number of alleles per variant.
    allele_ids
        Allele identifications.
    offset
        Variant offset.
    """

    id: DtypeLike
    rsid: DtypeLike
    chromosome: DtypeLike
    position: DtypeLike
    nalleles: DtypeLike
    allele_ids: DtypeLike
    offset: DtypeLike

    @property
    def size(self) -> int:
        """
        Number of variants.

        Returns
        -------
        Number of variants.
        """
        return self.id.shape[0]


@dataclass
class Partition:
    """
    Partition of variants.

    >>> import cbgen
    >>>
    >>> bgen = cbgen.bgen_file(cbgen.example.get("haplotypes.bgen"))
    >>> mf = cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile"))
    >>> part = mf.read_partition(0)
    >>> print(type(part))
    <class 'cbgen.typing.Partition'>
    >>> print(part.offset)
    0
    >>> print(type(part.variants))
    <class 'cbgen.typing.Variants'>
    >>> mf.close()
    >>> bgen.close()

    Attributes
    ----------
    offset
        Partition offset.
    variants
        Variants.
    """

    offset: int
    variants: Variants



from __future__ import annotations

from math import floor, sqrt
from pathlib import Path
from typing import Union

from numpy import empty, float32, float64, uint8, zeros

from cbgen.typing import CData, DtypeLike, Genotype

from ._ffi import ffi, lib

__all__ = ["bgen_file"]


class bgen_file:
    """
    BGEN file handler.

    >>> import cbgen
    >>>
    >>> bgen = cbgen.bgen_file(cbgen.example.get("haplotypes.bgen"))
    >>> print(bgen.nvariants)
    4
    >>> print(bgen.nsamples)
    4
    >>> print(bgen.contain_samples)
    True
    >>> print(bgen.read_samples())
    [b'sample_0' b'sample_1' b'sample_2' b'sample_3']
    >>> mf = cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile"))
    >>> part = mf.read_partition(0)
    >>> gt = bgen.read_genotype(part.variants.offset[0])
    >>> print(gt.probability)
    [[1. 0. 1. 0.]
     [0. 1. 1. 0.]
     [1. 0. 0. 1.]
     [0. 1. 0. 1.]]
    >>> mf.close()
    >>> bgen.close()

    Use `with`-statement context to guarantee file closing at the end.

    >>> with cbgen.bgen_file(cbgen.example.get("haplotypes.bgen")) as bgen:
    ...     print(bgen.nvariants)
    4

    Parameters
    ----------
    filepath
        BGEN file path.
    """

    def __init__(self, filepath: Union[str, Path]):
        self._filepath = Path(filepath)
        self._bgen_file: CData = ffi.NULL
        self._bgen_file = lib.bgen_file_open(bytes(self._filepath))
        if self._bgen_file == ffi.NULL:
            raise RuntimeError(f"Failed to open {filepath}.")

    @property
    def filepath(self) -> Path:
        """
        File path.

        Returns
        -------
        File path.
        """
        return self._filepath

    @property
    def nvariants(self) -> int:
        """
        Number of variants.

        Returns
        -------
        Number of variants.
        """
        return lib.bgen_file_nvariants(self._bgen_file)

    @property
    def nsamples(self) -> int:
        """
        Number of samples.

        Returns
        -------
        Number of samples.
        """
        return lib.bgen_file_nsamples(self._bgen_file)

    @property
    def contain_samples(self) -> bool:
        """
        Check if it contains samples.

        Returns
        -------
        ``True`` if it does contain samples; ``False`` otherwise.
        """
        return lib.bgen_file_contain_samples(self._bgen_file)

    def read_samples(self) -> DtypeLike:
        """
        Read samples.

        Returns
        -------
        Samples.

        Raises
        ------
        RuntimeError
            If samples are not stored or a file stream reading error occurs.
        """
        nsamples = self.nsamples
        bgen_samples: CData = lib.bgen_file_read_samples(self._bgen_file)
        if bgen_samples == ffi.NULL:
            raise RuntimeError("Could not fetch samples from the bgen file.")

        try:
            samples_max_len = ffi.new("uint32_t[]", 1)
            lib.read_samples_part1(bgen_samples, nsamples, samples_max_len)
            samples = zeros(nsamples, dtype=f"S{samples_max_len[0]}")
            lib.read_samples_part2(
                bgen_samples,
                nsamples,
                ffi.from_buffer("char[]", samples),
                samples_max_len[0],
            )
        finally:
            lib.bgen_samples_destroy(bgen_samples)

        return samples

    def create_metafile(self, filepath: Union[str, Path], verbose=False):
        """
        Create metafile file.

        Parameters
        ----------
        filepath
            File path.
        verbose
            ``True`` to show progress; ``False`` otherwise (default).
        """
        n = estimate_best_npartitions(self.nvariants)
        filepath = Path(filepath)

        mf = lib.bgen_metafile_create(self._bgen_file, bytes(filepath), n, verbose)
        if mf == ffi.NULL:
            raise RuntimeError(f"Error while creating metafile {filepath}.")

        lib.bgen_metafile_close(mf)

    def read_genotype(self, offset: int, precision: int = 64) -> Genotype:
        """
        Read genotype.

        Parameters
        ----------
        offset
            Variant offset.
        precision
            Probability precision in bits: 64 (default) or 32.

        Returns
        -------
        Genotype.

        Raises
        ------
        RuntimeError
            If invalid offset of or a file stream reading error occurs.
        """
        gt: CData = lib.bgen_file_open_genotype(self._bgen_file, offset)
        if gt == ffi.NULL:
            raise RuntimeError(f"Could not open genotype (offset {offset}).")

        if precision not in [64, 32]:
            raise ValueError("Precision should be either 64 or 32.")

        nsamples = self.nsamples
        ncombs = lib.bgen_genotype_ncombs(gt)
        err: int = 0
        if precision == 64:
            probs = empty((nsamples, ncombs), dtype=float64)
            err = lib.bgen_genotype_read64(gt, ffi.cast("double *", probs.ctypes.data))
        else:
            probs = empty((nsamples, ncombs), dtype=float32)
            err = lib.bgen_genotype_read32(gt, ffi.cast("float *", probs.ctypes.data))

        if err != 0:
            msg = f"Could not read genotype probabilities (offset {offset})."
            raise RuntimeError(msg)

        phased = lib.bgen_genotype_phased(gt)

        ploidy = empty(nsamples, dtype=uint8)
        lib.read_ploidy(gt, ffi.cast("uint8_t *", ploidy.ctypes.data), nsamples)

        missing = empty(nsamples, dtype=bool)
        lib.read_missing(gt, ffi.cast("bool *", missing.ctypes.data), nsamples)

        lib.bgen_genotype_close(gt)

        return Genotype(probs, phased, ploidy, missing)

    def read_probability(self, offset: int, precision: int = 64) -> DtypeLike:
        """
        Read genotype probability.

        Parameters
        ----------
        offset
            Variant offset.
        precision
            Probability precision in bits: 64 (default) or 32.

        Returns
        -------
        Probabilities.

        Raises
        ------
        RuntimeError
            If invalid offset of or a file stream reading error occurs.
        """
        gt: CData = lib.bgen_file_open_genotype(self._bgen_file, offset)
        if gt == ffi.NULL:
            raise RuntimeError(f"Could not open genotype (offset {offset}).")

        if precision not in [64, 32]:
            raise ValueError("Precision should be either 64 or 32.")

        nsamples = self.nsamples
        ncombs = lib.bgen_genotype_ncombs(gt)
        err: int = 0
        if precision == 64:
            probs = empty((nsamples, ncombs), dtype=float64)
            err = lib.bgen_genotype_read64(gt, ffi.cast("double *", probs.ctypes.data))
        else:
            probs = empty((nsamples, ncombs), dtype=float32)
            err = lib.bgen_genotype_read32(gt, ffi.cast("float *", probs.ctypes.data))

        if err != 0:
            msg = f"Could not read genotype probabilities (offset {offset})."
            raise RuntimeError(msg)

        lib.bgen_genotype_close(gt)

        return probs

    def close(self):
        """
        Close file stream.
        """
        if self._bgen_file != ffi.NULL:
            lib.bgen_file_close(self._bgen_file)
            self._bgen_file = ffi.NULL

    def __del__(self):
        self.close()

    def __enter__(self) -> bgen_file:
        return self

    def __exit__(self, *_):
        self.close()


def estimate_best_npartitions(nvariants: int) -> int:
    min_variants = 128
    m = max(min(min_variants, nvariants), floor(sqrt(nvariants)))
    return nvariants // m



import os
from pathlib import Path

from appdirs import user_cache_dir

__all__ = ["BGEN_CACHE_HOME"]

BGEN_CACHE_HOME = Path(
    os.environ.get("BGEN_CACHE_HOME", default=Path(user_cache_dir("bgen", "limix")))
)

(BGEN_CACHE_HOME / "test_data").mkdir(parents=True, exist_ok=True)



from importlib import import_module as _import_module

from . import example
from ._bgen_file import bgen_file
from ._bgen_metafile import bgen_metafile
from ._env import BGEN_CACHE_HOME
from ._testit import test

try:
    from ._ffi import ffi

    del ffi
except Exception as e:
    _ffi_err = """
It is likely caused by a broken installation of this package.
Please, make sure you have a C compiler and try to uninstall
and reinstall the package again."""

    raise RuntimeError(str(e) + _ffi_err)

try:
    __version__ = getattr(_import_module("cbgen._version"), "version", "x.x.x")
except ModuleNotFoundError:
    __version__ = "x.x.x"

__all__ = [
    "BGEN_CACHE_HOME",
    "__version__",
    "bgen_file",
    "bgen_metafile",
    "example",
    "test",
    "typing",
]



from __future__ import annotations

from pathlib import Path
from typing import Union

from numpy import empty, uint16, uint32, uint64, zeros

from cbgen.typing import CData, Partition, Variants

from ._ffi import ffi, lib

__all__ = ["bgen_metafile"]


class bgen_metafile:
    """
    BGEN metafile file handler.

    >>> import cbgen
    >>>
    >>> bgen = cbgen.bgen_file(cbgen.example.get("haplotypes.bgen"))
    >>> mf = cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile"))
    >>> print(mf.npartitions)
    1
    >>> print(mf.nvariants)
    4
    >>> print(mf.partition_size)
    4
    >>> part = mf.read_partition(0)
    >>> gt = bgen.read_genotype(part.variants.offset[0])
    >>> print(gt.probability)
    [[1. 0. 1. 0.]
     [0. 1. 1. 0.]
     [1. 0. 0. 1.]
     [0. 1. 0. 1.]]
    >>> mf.close()
    >>> bgen.close()

    Use `with`-statement context to guarantee file closing at the end.

    >>> with cbgen.bgen_metafile(cbgen.example.get("haplotypes.bgen.metafile")) as mf:
    ...     print(mf.npartitions)
    1

    Parameters
    ----------
    filepath
        BGEN metafile file path.

    Raises
    ------
    RuntimeError
        If a file stream reading error occurs.
    """

    def __init__(self, filepath: Union[str, Path]):
        self._filepath = Path(filepath)
        self._bgen_metafile: CData = ffi.NULL
        self._bgen_metafile = lib.bgen_metafile_open(bytes(self._filepath))
        if self._bgen_metafile == ffi.NULL:
            raise RuntimeError(f"Failed to open {filepath}.")

    @property
    def filepath(self) -> Path:
        """
        File path.

        Returns
        -------
        File path.
        """
        return self._filepath

    @property
    def npartitions(self) -> int:
        """
        Number of partitions.

        Returns
        -------
        Number of partitions.
        """
        return lib.bgen_metafile_npartitions(self._bgen_metafile)

    @property
    def nvariants(self) -> int:
        """
        Number of variants.

        Returns
        -------
        Number of variants.
        """
        return lib.bgen_metafile_nvariants(self._bgen_metafile)

    @property
    def partition_size(self) -> int:
        """
        Number of variants per partition.

        The last partition might have less variants than the partition size.
        Every other partition is guaranteed to have ``partition_size``
        variants.

        Returns
        -------
        Partition size.
        """
        return ceildiv(self.nvariants, self.npartitions)

    def read_partition(self, index: int) -> Partition:
        """
        Read partition.

        Parameters
        ----------
        index
            Partition index.

        Returns
        -------
        Partition.

        Raises
        ------
        RuntimeError
            If index is invalid or a file stream reading error occurs.
        """
        partition = lib.bgen_metafile_read_partition(self._bgen_metafile, index)
        if partition == ffi.NULL:
            raise RuntimeError(f"Could not read partition {partition}.")

        nvariants = lib.bgen_partition_nvariants(partition)

        position = empty(nvariants, dtype=uint32)
        nalleles = empty(nvariants, dtype=uint16)
        var_offset = empty(nvariants, dtype=uint64)
        vid_max_len = ffi.new("uint32_t[]", 1)
        rsid_max_len = ffi.new("uint32_t[]", 1)
        chrom_max_len = ffi.new("uint32_t[]", 1)
        allele_ids_max_len = ffi.new("uint32_t[]", 1)

        position_ptr = ffi.cast("uint32_t *", ffi.from_buffer(position))
        nalleles_ptr = ffi.cast("uint16_t *", ffi.from_buffer(nalleles))
        offset_ptr = ffi.cast("uint64_t *", ffi.from_buffer(var_offset))
        lib.read_partition_part1(
            partition,
            position_ptr,
            nalleles_ptr,
            offset_ptr,
            vid_max_len,
            rsid_max_len,
            chrom_max_len,
            allele_ids_max_len,
        )

        vid = zeros(nvariants, dtype=f"S{vid_max_len[0]}")
        rsid = zeros(nvariants, dtype=f"S{rsid_max_len[0]}")
        chrom = zeros(nvariants, dtype=f"S{chrom_max_len[0]}")
        allele_ids = zeros(nvariants, dtype=f"S{allele_ids_max_len[0]}")

        lib.read_partition_part2(
            partition,
            ffi.from_buffer("char[]", vid),
            vid_max_len[0],
            ffi.from_buffer("char[]", rsid),
            rsid_max_len[0],
            ffi.from_buffer("char[]", chrom),
            chrom_max_len[0],
            ffi.from_buffer("char[]", allele_ids),
            allele_ids_max_len[0],
        )
        lib.bgen_partition_destroy(partition)

        part_offset = self.partition_size * index
        v = Variants(vid, rsid, chrom, position, nalleles, allele_ids, var_offset)
        return Partition(part_offset, v)

    def close(self):
        """
        Close file stream.
        """
        if self._bgen_metafile != ffi.NULL:
            lib.bgen_metafile_close(self._bgen_metafile)
            self._bgen_metafile = ffi.NULL

    def __del__(self):
        self.close()

    def __enter__(self) -> bgen_metafile:
        return self

    def __exit__(self, *_):
        self.close()


def ceildiv(a: int, b: int) -> int:
    return -(-a // b)



import logging
from pathlib import Path

import pooch

from ._env import BGEN_CACHE_HOME

__all__ = ["get"]

pooch.get_logger().setLevel(logging.ERROR)

goodboy = pooch.create(
    path=BGEN_CACHE_HOME / "test_data",
    base_url="https://bgen-examples.s3.amazonaws.com/",
    registry={
        "complex.23bits.no.samples.bgen": "25d30a4e489da1aeb05f9893af98e8bf3b09d74db2982bf1828f8c8565886fc6",
        "haplotypes.bgen": "84e0b59efcc83c7c305cf5446e5dc26b49b15aeb4157a9eb36451376ce3efe4c",
        "haplotypes.bgen.metadata.corrupted": "8f55628770c1ae8155c1ced2463f15df80d32bc272a470bb1d6b68225e1604b1",
        "haplotypes.bgen.metafile": "7e8b13aed04e2166649dde2c1c44aa4239f158aef7cf06d69c51113b9a2ae175",
        "wrong.metadata": "f746345605150076f3234fbeea7c52e86bf95c9329b2f08e1e3e92a7918b98fb",
        "merged_487400x220000.bgen": "blake2b:2bf8043907eff9c00021dd044d80d63873f465898b8732c8bcbb533ca5fcd63875aa54e8103768631610f76c4c82577bc7c15f7bea90bb1ab1906b12ceddf56e",
        "merged_487400x2420000.bgen": "81aecfab787bee1cb7f1d0d21f2465c581a4db78011d8b0f0f73c868e17ec888",
        "merged_487400x4840000.bgen": "5ef82f92a001615c93bbb317a9fd2329272370c6d481405d4f8f0a2b7fddf68b",
    },
)


def get(filename: str) -> Path:
    """
    Get file path to an example.

    Recognized file names:

    - ``complex.23bits.no.samples.bgen``
    - ``haplotypes.bgen``
    - ``haplotypes.bgen.metadata.corrupted``
    - ``haplotypes.bgen.metafile``
    - ``wrong.metadata``
    - ``merged_487400x220000.bgen``
    - ``merged_487400x2420000.bgen``
    - ``merged_487400x4840000.bgen``

    Parameters
    ----------
    filename
        File name to fetch.

    Returns
    -------
    File path.
    """
    return Path(goodboy.fetch(filename))



from pathlib import Path

import pytest
from numpy import isnan, nan, nansum
from numpy.testing import assert_allclose, assert_array_equal

from cbgen import bgen_file, bgen_metafile, example


@pytest.mark.slow
def test_cbgen_large1(tmp_path):
    filepath = example.get("merged_487400x220000.bgen")
    mfilepath = tmp_path / f"{filepath.name}.metafile"

    bgen = bgen_file(filepath)
    assert bgen.nvariants == 220000
    assert bgen.nsamples == 487400
    assert not bgen.contain_samples

    with pytest.raises(RuntimeError):
        bgen.read_samples()

    bgen.create_metafile(mfilepath, verbose=True)

    mf = bgen_metafile(mfilepath)

    assert mf.filepath.name == mfilepath.name
    assert mf.npartitions == 469
    assert mf.nvariants == 220000
    assert mf.partition_size == 470

    part = mf.read_partition(5)

    assert part.variants.size == 470
    assert part.variants.id[0] == b"sid_1_2350"
    assert part.variants.rsid[0] == b"sid_1_2350"
    assert part.variants.chromosome[0] == b"1"
    assert part.variants.position[0] == 2351
    assert part.variants.nalleles[0] == 2
    assert part.variants.allele_ids[0] == b"A,C"
    voff = part.variants.offset[0]
    gt = bgen.read_genotype(voff)
    assert_allclose(gt.probability.shape, (487400, 3))
    assert_allclose(nansum(gt.probability, 0), [475743.0, 0.0, 0.0])
    assert_allclose(isnan(gt.probability).sum(0), [11657, 11657, 11657])
    assert not gt.phased
    assert_allclose(gt.ploidy.sum(), 974800)
    assert_allclose(gt.missing.sum(), 11657)

    mf.close()
    bgen.close()


def test_cbgen_nonexistent_bgen_file():
    with pytest.raises(RuntimeError):
        bgen_file("/Fmw/DiKel")


def test_cbgen_error_create_metafile():
    filepath = example.get("haplotypes.bgen")
    mfilepath = "/DmEkq/WkhDu/bla.metafile"

    bgen = bgen_file(filepath)

    with pytest.raises(RuntimeError):
        bgen.create_metafile(mfilepath, verbose=False)


def test_cbgen_invalid_metafile():
    mfilepath = example.get("wrong.metadata")
    with pytest.raises(RuntimeError):
        bgen_metafile(mfilepath)

    mfilepath = example.get("haplotypes.bgen.metadata.corrupted")
    with pytest.raises(RuntimeError):
        bgen_metafile(mfilepath)


def test_cbgen_phased_genotype(tmp_path):
    filepath = example.get("haplotypes.bgen")
    mfilepath = tmp_path / f"{filepath.name}.metafile"

    bgen = bgen_file(filepath)
    assert bgen.filepath.name == "haplotypes.bgen"
    assert bgen.nvariants == 4
    assert bgen.nsamples == 4
    assert bgen.contain_samples
    samples = bgen.read_samples()
    assert_array_equal(samples, [b"sample_0", b"sample_1", b"sample_2", b"sample_3"])

    bgen.create_metafile(mfilepath, verbose=False)

    mf = bgen_metafile(mfilepath)

    assert mf.filepath.name == mfilepath.name
    assert mf.npartitions == 1
    assert mf.nvariants == 4
    assert mf.partition_size == 4

    part = mf.read_partition(0)

    assert part.variants.size == 4
    assert part.variants.id[0] == b"SNP1"
    assert part.variants.rsid[0] == b"RS1"
    assert part.variants.chromosome[0] == b"1"
    assert part.variants.position[0] == 1
    assert part.variants.nalleles[0] == 2
    assert part.variants.allele_ids[0] == b"A,G"
    voff = part.variants.offset[0]
    gt = bgen.read_genotype(voff)
    assert_allclose(
        gt.probability,
        [
            [1.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 1.0],
        ],
    )
    assert gt.phased
    assert_allclose(gt.ploidy, [2, 2, 2, 2])
    assert_allclose(gt.missing, [False, False, False, False])

    gt = bgen.read_genotype(voff, 32)
    assert_allclose(
        gt.probability,
        [
            [1.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 1.0],
        ],
    )

    assert part.variants.id[3] == b"SNP4"
    assert part.variants.rsid[3] == b"RS4"
    assert part.variants.chromosome[3] == b"1"
    assert part.variants.position[3] == 4
    assert part.variants.nalleles[3] == 2
    assert part.variants.allele_ids[3] == b"A,G"
    voff = part.variants.offset[3]
    gt = bgen.read_genotype(voff)
    assert_allclose(
        gt.probability,
        [
            [0.0, 1.0, 0.0, 1.0],
            [1.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [1.0, 0.0, 0.0, 1.0],
        ],
    )
    assert gt.phased
    assert_allclose(gt.ploidy, [2, 2, 2, 2])
    assert_allclose(gt.missing, [False, False, False, False])

    gt = bgen.read_genotype(voff, 32)
    assert_allclose(
        gt.probability,
        [
            [0.0, 1.0, 0.0, 1.0],
            [1.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [1.0, 0.0, 0.0, 1.0],
        ],
    )

    mf.close()
    bgen.close()


def test_cbgen_complex_unphased(tmp_path: Path):
    filepath = example.get("complex.23bits.no.samples.bgen")
    mfilepath = tmp_path / f"{filepath.name}.metafile"
    with bgen_file(filepath) as bgen:
        assert bgen.filepath.name == "complex.23bits.no.samples.bgen"
        assert bgen.nvariants == 10
        assert bgen.nsamples == 4
        assert not bgen.contain_samples
        with pytest.raises(RuntimeError):
            bgen.read_samples()
        bgen.create_metafile(mfilepath, verbose=False)

    with bgen_metafile(mfilepath) as mf:
        assert mf.filepath.name == mfilepath.name
        assert mf.npartitions == 1
        assert mf.nvariants == 10
        assert mf.partition_size == 10

        part = mf.read_partition(0)

        assert part.variants.id[0] == b""
        assert part.variants.rsid[0] == b"V1"
        assert part.variants.chromosome[0] == b"01"
        assert part.variants.position[0] == 1
        assert part.variants.nalleles[0] == 2
        assert part.variants.allele_ids[0] == b"A,G"

        with bgen_file(filepath) as bgen:
            voff = part.variants.offset[0]
            gt = bgen.read_genotype(voff)
            assert_allclose(
                gt.probability,
                [[1.0, 0.0, nan], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
            )
            assert_allclose(
                bgen.read_probability(voff),
                [[1.0, 0.0, nan], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
            )
            assert not gt.phased
            assert_allclose(gt.ploidy, [1, 2, 2, 2])
            assert_allclose(gt.missing, [False, False, False, False])

            gt = bgen.read_genotype(voff, 32)
            assert_allclose(
                gt.probability,
                [[1.0, 0.0, nan], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
            )

            voff = part.variants.offset[-1]
            gt = bgen.read_genotype(voff)
            assert_allclose(
                gt.probability,
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                ],
            )
            assert not gt.phased
            assert_allclose(gt.ploidy, [4, 4, 4, 4])
            assert_allclose(gt.missing, [False, False, False, False])

            gt = bgen.read_genotype(voff, 32)
            assert_allclose(
                gt.probability,
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                ],
            )

            with pytest.raises(ValueError):
                bgen.read_genotype(voff, 12)

            valid_offsets = set(list(part.variants.offset))
            all_offsets = set(list(range(0, int(max(valid_offsets)) + 1)))
            invalid_offsets = all_offsets - valid_offsets

            for offset in list(invalid_offsets):
                with pytest.raises(RuntimeError):
                    bgen.read_genotype(offset)

        with pytest.raises(RuntimeError):
            part = mf.read_partition(1)
